;; Generated package description from elisp-benchmarks.el  -*- no-byte-compile: t -*-
(define-package "elisp-benchmarks" "1.16.0.20240708.114026" "elisp benchmarks collection" 'nil :commit "6446f6abf91f81e7160295eecf3cd90821bab9d1" :url "https://elpa.gnu.org/packages/elisp-benchmarks.html" :authors '(("Andrea Corallo" . "acorallo@gnu.org")) :maintainer '("Andrea Corallo" . "acorallo@gnu.org") :keywords '("languages" "lisp"))
